#!/usr/bin/env Rscript

# Cluster distance matrix version 0.0.1
# Copyright (C) 2018 onwards Carlus Deneke


# supress warnings
options(warn=-1)

suppressPackageStartupMessages(library(optparse))
suppressPackageStartupMessages(library("ape"))
suppressPackageStartupMessages(library(dendextend))
suppressPackageStartupMessages(library(magrittr))


options(warn=0)

# option parsing #

parser <- OptionParser()

parser <- add_option(parser, 
                     opt_str = c("-i", "--input"), 
                     type = "character",
                     dest = 'input.file',
                     help="Distance matrix (tsv) Input file (required)"
)

parser <- add_option(parser, 
                     opt_str = c("-o", "--outputdirectory"), 
                     type = "character",
                     dest = 'output.dir',
                     default=NA,
                     help="Output directory (optional, default is the same as the input directory)."
)

parser <- add_option(parser, 
                     opt_str = c("-c", "--cutoff_values"), 
                     type="character", 
                     default="5,10,20,50,100,200,1000",
                     dest = 'thresholds',
                     help="A comma separated set of cutoff values (Default: 5,10,20,50,100,200,1000)"
)

parser <- add_option(parser,
                     opt_str = c( "--header"),
                     action = "store_true",
                     dest = 'header',
                     help="TRUE or FALSE (the default). When true, the first line in the distance matrix is the header"
)



parser <- add_option(parser,
                     opt_str = c( "--tsv"),
                     action = "store_true",
                     dest = 'tsv',
                     help="TRUE or FALSE (the default). When true, the distance matrix is tab seperated (else space seperated)"
)



parser <- add_option(parser,
                     opt_str = c("-f", "--force"),
                     action = "store_true",
                     dest = 'force',
                     help="TRUE or FALSE (the default). When true, existing output files will be overwritten"
)

opt = parse_args(parser)



if (length(opt$input.file)==1) {
  distance_matrix_file = opt$input.file
} else {
  stop("Input file parameter must be supplied via -i or --input. See script usage (--help)")
}


if(isTRUE(opt$force)){
  Force <- T
} else {
  Force <- F
}



if(isTRUE(opt$header)){
  header <- T
} else {
  header <- F
}



if(isTRUE(opt$tsv)){
  sep <- "\t"
} else {
  sep <- " "
}

if(!is.na(opt$output.dir)){
  outdir <- opt$output.dir  
}else {
  outdir <- getwd()
}

# check if dir exists
if(!dir.exists(outdir)){
  warning(paste("Directory",outdir,"does not exist"))
  dir.create(outdir)
} 

# defining output files
out_cluster <- file.path(outdir,"cluster_groups.tsv")
out_address <- file.path(outdir,"mlst_address.tsv")
out_pdf <- file.path(outdir,"clustering_thresholds.pdf")
out_pdf_2 <- file.path(outdir,"clustering_sizes.pdf")


# parse cutoff values:

cut_threshold_range_raw <- opt$thresholds
cut_threshold_range_raw <- trimws(cut_threshold_range_raw, which = c("both"))
cut_threshold_range <- as.numeric(strsplit(cut_threshold_range_raw,"[ ,;]+")[[1]])


# initial messages -----------------------------------


message(paste(rep("=",50),collapse = "="))



message("\nRunning Clustering_DistanceMatrix.R on",date(),"\n\n")


message(paste("distance_matrix_file:",distance_matrix_file))
message(paste("cut_threshold_range:",cut_threshold_range_raw))
message(paste("outdir:",outdir))
message(paste("force:",Force))

message(paste("Number of supplied cutoff values:",length(cut_threshold_range)))
for (i in 1:length(cut_threshold_range)){
  message(paste0("cutoff value ",i,": ",cut_threshold_range[i]))
}

message("\n",paste(rep("=",50),collapse = "="),"\n")

# check if output files exist --------------
if(file.exists(out_cluster) & !isTRUE(Force) ){
  stop("File",out_cluster,"already exists. Use flag --force to overwrite")
}

if(file.exists(out_address) & !isTRUE(Force) ){
  stop("File",out_address,"already exists. Use flag --force to overwrite")
}

if(file.exists(out_pdf) & !isTRUE(Force) ){
  stop("File",out_pdf,"already exists. Use flag --force to overwrite")
}

if(file.exists(out_pdf_2) & !isTRUE(Force) ){
  stop("File",out_pdf_2,"already exists. Use flag --force to overwrite")
}

if(any(is.na(cut_threshold_range))){
  stop("Values in --cutoff_values not properly defined:",paste(cut_threshold_range,collapse = ","))
}



# function ----------------------------------

read_distance_matrix <- function(file, sep = " ", header = FALSE){
  if(!file.exists(file)) {stop(paste("File",file,"does Not exist")); return(NA)}
  distance_matrix <- read.delim(file, sep=sep, stringsAsFactors = F, header = header)
  rownames(distance_matrix) <- sub(".fasta","",distance_matrix[,1])
  distance_matrix <- distance_matrix[,-1]
  colnames(distance_matrix) <- rownames(distance_matrix)
  distance_matrix <- as.matrix(distance_matrix[order(colnames(distance_matrix)),order(colnames(distance_matrix))])
  return(distance_matrix)
}

# read in data -----------------------------------


distance_matrix <- distance_matrix_file %>% read_distance_matrix (sep = sep, header = header)

dendro <- distance_matrix %>% as.dist %>% hclust(., method = "single") %>% as.dendrogram 


# plot cut-off values and clustering ----------------

pdf(file = out_pdf, title = "Clustering and cutoffs", width = 32, height = 18)
plot(dendro, horiz = T)
abline(v= cut_threshold_range, col = rainbow(length(cut_threshold_range)), lwd = 2)

graphics.off()

message("Output written to ",out_pdf)

## plot number of clusters per threshold

subcluster_counts <- sapply(clustering_threshold_range, function(cut_threshold) {
  length(unique(dendextend::cutree(dendro, h = cut_threshold)))  
})
names(subcluster_counts) <- clustering_threshold_range

pdf(file = out_pdf_2, title = "Cluster sizes vs. cutoffs", width = 32, height = 18)
plot(names(subcluster_counts),subcluster_counts, xlab="Cluster threshold", ylab = "Number of clusters", type = "b")
graphics.off()

message("Output written to ",out_pdf_2)


# compute cluster memberships and addresses for all cutoffs
mlst_address_matrix <- do.call(rbind,lapply(sort(cut_threshold_range,decreasing = T), function(x){
  dendextend::cutree(dendro, h = x)  
}))
rownames(mlst_address_matrix) <- paste0("CT_",as.character(sort(cut_threshold_range, decreasing = T)) )
mlst_address_matrix <-t(mlst_address_matrix)

# convert to address
mlst_address <- apply(mlst_address_matrix,1,function(x) paste(x, collapse = "."))

# export
write.table(cbind(names(mlst_address),mlst_address), file = out_address, sep = "\t", quote = F, col.names = c("ID","Address"), row.names = F)
write.table(cbind(rownames(mlst_address_matrix),mlst_address_matrix), file = out_cluster, sep = "\t", quote = F,col.names = c("ID",colnames(mlst_address_matrix)), row.names = F)

# finalize
message("Output written to ",out_address)
message("Output written to ",out_cluster)





