
# function for reading distance matrix, as provided by grapetree or snp-dist
read_distance_matrix <- function(file, sep = " ", header = FALSE){
  if(!file.exists(file)) {stop(paste("File",file,"does Not exist")); return(NA)}
  distance_matrix <- read.delim(file, sep=sep, stringsAsFactors = F, header = header)
  rownames(distance_matrix) <- sub(".fasta","",distance_matrix[,1])
  distance_matrix <- distance_matrix[,-1]
  colnames(distance_matrix) <- rownames(distance_matrix)
  distance_matrix <- as.matrix(distance_matrix[order(colnames(distance_matrix)),order(colnames(distance_matrix))])
  return(distance_matrix)
}

# utility function for ordering correlation matrix
reorder_cormat <- function(cormat){
  dd <- as.dist(cormat)
  hc <- hclust(dd)
  cormat <-cormat[hc$order, hc$order]
}

# utility function for matrix operations
get_upper_tri <- function(cormat){
  cormat[lower.tri(cormat)]<- NA
  return(cormat)
}

# function for plotting
plot_distance_matrix <- function(distance_matrix2, param_color_thr = c(0,1,5,10,20), param_size = 3, param_legendsize = 12, label_color = "black"){
  
  # remove diagonal elements
  diag(distance_matrix2) <- NA
  
  # Reorder the correlation matrix
  distance_matrix2 <- reorder_cormat(distance_matrix2)
  
  # melt matrix
  distance_matrix.m <- reshape2::melt(distance_matrix2, na.rm = TRUE, value.name="Distance",varnames=c("Sample1","Sample2"), id.vars = "id")
  
  # reorder the label
  label_color <- label_color[match(rownames(distance_matrix2),names(label_color))]
  label_color[is.na(label_color)] <- "black"
  
  p1 <- ggplot(distance_matrix.m, aes(Sample1, Sample2)) +
    geom_tile(aes(fill = Distance), colour = "white", show.legend = F) + 
    geom_text(aes(label = round(Distance, 1)), size = param_size) +
    scale_fill_gradientn(colours = c("white", "yellow", "orange", "red"), values = scales::rescale(param_color_thr)) +
    theme(axis.text.x = element_text(angle = 90, vjust = 0.5, size = param_legendsize, hjust = 1, color = label_color),  axis.text.y = element_text(size = param_legendsize, color = label_color), axis.title.x = element_blank(),
          axis.title.y = element_blank()) + coord_fixed()
  
  
  return(p1)
  
}

# function that cuts dendrogram at threshold and returns a table with sample to cluster number association
Get_samples2cluster <- function(dendro, cut_threshold, removeOrphans=F, renameOrphans=T){
  
  # check
  if(!is.dendrogram(dendro)) stop("Provided input is not of type dendrogram")
  
  # get clusters at threshold
  # >>> NEW
  dendro_cut <- dendextend::cutree(dendro, h = cut_threshold)  
  samples2cluster <- data.frame(sample=names(dendro_cut),cluster_number=as.character(dendro_cut), stringsAsFactors = F)
  # rename orphans
  cluster_size_count <- table(samples2cluster$cluster_number)
  numbers_orphanclustersmytable <- names(cluster_size_count[which(cluster_size_count == 1)])
  samples2cluster$cluster_number <- ifelse(samples2cluster$cluster_number %in% numbers_orphanclustersmytable, paste0("Orphan_",samples2cluster$cluster_number),samples2cluster$cluster_number)

  # remove orphans
  if(removeOrphans) {
    samples2cluster <- samples2cluster[!grepl("Orphan",samples2cluster$cluster_number),]
  }
  return(samples2cluster)
}

# find conflicts
check_clusternaming_conflicts <- function(clusternumber2name){
  if(any(table(clusternumber2name$cluster_number) > 1)){
    # 2 or more cluster_numbers point to the same cluster_name
    #exception1 <- which(table(clusternumber2name$cluster_number) > 1)
    exception1 <- names(which(table(clusternumber2name$cluster_number) > 1))
    # debugging
    exception1_table <- clusternumber2name[clusternumber2name$cluster_number %in% exception1,]
    print(exception1_table)
    warning("Unexpected behaviour: Two cluster names point to a single Cluster number.\n Clusters have merged!\nRelated cluster numbers are ",paste(exception1,collapse = ","))
    #action handled in function Fix_merge_conflicts: combine cluster_names; re-assign related samples to new cluster name; Avoid reassignment of retired cluster names
  } else if (any(table(clusternumber2name$cluster_name) > 1)){
    # 2 or more cluster_names point to the same cluster_number
    exception2 <- names(which(table(clusternumber2name$cluster_name) > 1))
    exception2_table <- clusternumber2name[clusternumber2name$cluster_name %in% exception2,]
    print(exception2_table)
    stop("Unexpected behaviour: Two cluster numbers point to a single Cluster name.\n Clusters have split!\nRelated cluster numbers are ",paste(exception2,collapse = ","))
    # TODO action: Re-assign samples to two different cluster_names; e.g. cluster_name_split1 and cluster_name_split2
  }
}



# Fix merge conflict by concatenating merge conflict cluster names with ":"
Fix_merge_conflicts <- function(clusternumber2name,samples2clustername, verbose=T){
  
  conflict_cluster_number_array <- names(which(table(clusternumber2name$cluster_number) > 1))
  
  if(length(conflict_cluster_number_array) == 0) {
    if(verbose) print("No merge conflicts")
    return(list(clusternumber2name=clusternumber2name,samples2clustername=samples2clustername))
  }
  
  # for each conflict
  for(conflict_cluster_number in conflict_cluster_number_array){
    if(verbose) print(paste("Fixing merge conflict for cluster number",conflict_cluster_number))
    
    # find new name
    cluster_name_merged <- paste(sort(clusternumber2name$cluster_name[clusternumber2name$cluster_number == conflict_cluster_number]), collapse = ":")
    
    # 2a. rename sample2cluster
    samples2clustername$cluster_name[samples2clustername$cluster_number == conflict_cluster_number] <- cluster_name_merged
    # 2b. rename clusternumber2name
    clusternumber2name$cluster_name[clusternumber2name$cluster_number == conflict_cluster_number] <- cluster_name_merged
  }
  
  # make unique
  clusternumber2name <- clusternumber2name[!duplicated(clusternumber2name$cluster_name),]
  
  if(verbose) print("All merge conflicts solved")
  return(list(clusternumber2name=clusternumber2name,samples2clustername=samples2clustername))
  
}



# get new cluster name
Get_newclustername <- function(clusternumber2name,potential_clusternames){
  # find new cluster name
  #used_names <- unique(clusternumber2name$cluster_name)
  used_names <- unique(unlist(sapply(clusternumber2name$cluster_name, function(x) strsplit(x,":")[[1]]))) #also subtract merged names
  available_clusternames <- setdiff(potential_clusternames, used_names)
  
  if(length(available_clusternames) == 0) {
    new_clustername <- paste0("novel_",digest::digest(Sys.time(),algo = "xxhash32"))
    warning(paste("No more available_clusternames cluster names, creating a random cluster name:",new_clustername))
  } else {
    new_clustername <- available_clusternames[1]
  }
  return(new_clustername)
}


# assign novel cluster numbers to samples

update_samplenames <- function(samples2clustername, clusternumber2name, verbose=F, potential_clusternames){
  
  unmatched_samples <- is.na(samples2clustername$cluster_name)
  samples2clustername_unmatched <- samples2clustername[unmatched_samples,]
  
  if(sum(unmatched_samples) == 0){
    if(verbose) print("No update required")
    return(list(samples2clustername=samples2clustername,clusternumber2name=clusternumber2name))
  }
  
  if(verbose) print(paste("Updating",sum(unmatched_samples),"samples"))
  
  for(i in 1:nrow(samples2clustername_unmatched)){
    samples2clustername_unmatched$sample[i]
    my_cluster_number <- samples2clustername_unmatched$cluster_number[i]
    
    if(my_cluster_number %in% clusternumber2name$cluster_number){
      # cluster name already exists
      print("cluster name already exists")
      my_cluster_name <- clusternumber2name$cluster_name[match(my_cluster_number,clusternumber2name$cluster_number)]
      print(my_cluster_name)
    } else {
      # cluster name does NOT yes exist
      print("Novel cluster name")
      # assign new clusternames for new cluster_numbers
      my_cluster_name <- Get_newclustername(clusternumber2name = clusternumber2name, potential_clusternames = potential_clusternames)
      # update clusternumber2name
      clusternumber2name <- clusternumber2name %>% add_row(cluster_number=my_cluster_number,cluster_name = my_cluster_name)
    }
    # assign name to temporary data frame of unmatched samples
    samples2clustername_unmatched$cluster_name[i] <- my_cluster_name
  }
  
  # update names of unmatched samples
  samples2clustername[unmatched_samples,] <- samples2clustername_unmatched
  
  return(list(samples2clustername=samples2clustername,clusternumber2name=clusternumber2name))
}


# -----------

make_color_coding <- function(subclusters, i){
  
  subclusters_uniq <- sort(unique(subclusters[,i]))
  
  if(length(subclusters_uniq) == 1){
    subclusters_color <- "gray"
  } else if(length(subclusters_uniq) == 2){
    subclusters_color <- c("red","blue")
  } else if (length(subclusters_uniq) > 12) { # handle cases where more clusters than palette by adding second palette
    subclusters_color_1 <- brewer.pal(12, "Set3") #,"RdBu"  
    subclusters_color_2 <- brewer.pal(length(subclusters_uniq)-12, "Set1")
    subclusters_color <- c(subclusters_color_1,subclusters_color_2)
  } else if (length(subclusters_uniq) > 2) {
    subclusters_color <- brewer.pal(length(subclusters_uniq), "Set3") #,"RdBu"  
  } else {
    subclusters_color <- "black"
  }
  
  names(subclusters_color) <- subclusters_uniq[1:length(subclusters_color)]
  
  # table of sample and color
  subclusters_color_table <- subclusters_color[subclusters[,i]]
  names(subclusters_color_table) <- subclusters[,1]
  
  return(subclusters_color_table)
}
